/****************************************************************************
*
*    The MIT License (MIT)
*
*    Copyright (c) 2014 - 2022 Vivante Corporation
*
*    Permission is hereby granted, free of charge, to any person obtaining a
*    copy of this software and associated documentation files (the "Software"),
*    to deal in the Software without restriction, including without limitation
*    the rights to use, copy, modify, merge, publish, distribute, sublicense,
*    and/or sell copies of the Software, and to permit persons to whom the
*    Software is furnished to do so, subject to the following conditions:
*
*    The above copyright notice and this permission notice shall be included in
*    all copies or substantial portions of the Software.
*
*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
*    DEALINGS IN THE SOFTWARE.
*
*****************************************************************************
*
*    The GPL License (GPL)
*
*    Copyright (C) 2014 - 2022 Vivante Corporation
*
*    This program is free software; you can redistribute it and/or
*    modify it under the terms of the GNU General Public License
*    as published by the Free Software Foundation; either version 2
*    of the License, or (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software Foundation,
*    Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*
*****************************************************************************
*
*    Note: This software is released under dual MIT and GPL licenses. A
*    recipient may use this file under the terms of either the MIT license or
*    GPL License. If you wish to use only one license not the other, you can
*    indicate your decision by deleting one of the above license notices in your
*    version of this file.
*
*****************************************************************************/


#include <linux/device.h>
#include <linux/slab.h>
#include <linux/miscdevice.h>
#include <linux/uaccess.h>
#include <linux/module.h>
#include "common.h"

#include "gc_hal_kernel_linux.h"
#include "gc_hal_driver.h"
#include "gc_hal_kernel_parameter.h"
#include <linux/platform_device.h>
#include "sn_osal.h"
#include "memory_osal.h"

/* Zone used for header/footer. */
#define _GC_OBJ_ZONE gcvZONE_DRIVER

// MODULE_DESCRIPTION("Vivante Graphics Driver");
// MODULE_LICENSE("Dual MIT/GPL");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0)
// MODULE_IMPORT_NS(VFS_internal_I_am_really_a_filesystem_and_am_NOT_a_driver);
#endif

static uint regOffsets[gcdGLOBAL_CORE_COUNT] =
{
    [0 ... gcdGLOBAL_CORE_COUNT - 1] = 0,
};

static int gpu3DMinClock = 1;
static int contiguousRequested;
static ulong bankSize;

static void
_InitModuleParam(gcsMODULE_PARAMETERS *ModuleParam)
{
    gctUINT i, j;
    gcsMODULE_PARAMETERS *p = ModuleParam;

    for (i = 0; i < gcdDEVICE_COUNT; i++)
        p->devices[i] = gcvNULL;

    for (i = 0; i < gcdGLOBAL_CORE_COUNT; i++) {
        p->irqs[i] = irqs[i];
        if (irqs[i] != -1) {
            p->registerBases[i] = registerBases[i];
            p->registerSizes[i] = registerSizes[i];
        }
#if USE_LINUX_PCIE
        p->bars[i] = bars[i];
        p->regOffsets[i] = regOffsets[i];
#endif
        p->registerBasesMapped[i] = gcvNULL;

        for (j = 0; j < gcvSRAM_INTER_COUNT; j++) {
            p->sRAMBases[i][j] = sRAMBases[i * gcvSRAM_INTER_COUNT + j];
            p->sRAMSizes[i][j] = sRAMSizes[i * gcvSRAM_INTER_COUNT + j];
        }
    }

    for (i = 0; i < gcvCORE_COUNT; i++)
        p->chipIDs[i] = chipIDs[i];

    for (i = 0; i < gcdGLOBAL_2D_COUNT; i++) {
        p->irq2Ds[i] = irq2Ds[i];

        if (irq2Ds[i] != -1) {
            p->register2DBases[i] = register2DBases[i];
            p->register2DSizes[i] = register2DSizes[i];
        }

#if USE_LINUX_PCIE
        p->bar2Ds[i] = bar2Ds[i];
        p->reg2DOffsets[i] = reg2DOffsets[i];
#endif
    }

    if (irqLine != -1) {
        p->irqs[gcvCORE_MAJOR] = irqLine;
        p->registerBases[gcvCORE_MAJOR] = registerMemBase;
        p->registerSizes[gcvCORE_MAJOR] = registerMemSize;
        /* Check legacy style. */
#if USE_LINUX_PCIE
        if (bar != -1) {
            if (p->bars[gcvCORE_MAJOR] == -1)
                p->bars[gcvCORE_MAJOR] = bar;
        }
#endif
    }

    if (irqLine2D != -1) {
# if S1_GPU
        p->irq2Ds[0] = irqLine2D;
        p->register2DBases[0] = registerMemBase2D;
        p->register2DSizes[0] = registerMemSize2D;
#if USE_LINUX_PCIE
        if (bar2D != -1) {
            if (p->bar2Ds[0] == -1)
                p->bar2Ds[0] = bar2D;
        }
#endif
#elif !S1_GPU && S2_GPU
    p->irq2Ds[1] = irqLine2D;
        p->register2DBases[1] = registerMemBase2D;
        p->register2DSizes[1] = registerMemSize2D;
#if USE_LINUX_PCIE
        if (bar2D != -1) {
            if (p->bar2Ds[1] == -1)
                p->bar2Ds[1] = bar2D;
        }
#endif
#endif
    }

    p->irqVG = irqLineVG;

    if (irqLineVG != -1) {
        p->registerVGBase = registerMemBaseVG;
        p->registerVGSize = registerMemSizeVG;
    }

#if USE_LINUX_PCIE
    p->barVG = barVG;
    p->regVGOffset = regVGOffset;
#endif

#if gcdDEC_ENABLE_AHB
    if (registerMemBaseDEC300 && registerMemSizeDEC300) {
        p->registerBases[gcvCORE_DEC] = registerMemBaseDEC300;
        p->registerSizes[gcvCORE_DEC] = registerMemSizeDEC300;
    }
#endif

    for (i = 0; i < gcdSYSTEM_RESERVE_COUNT; i++) {
        p->contiguousBases[i] = contiguousBases[i];
        p->contiguousSizes[i] = contiguousSizes[i];
    }

    p->contiguousBase = contiguousBase;
    p->contiguousSizes[0] = contiguousSize;
    p->contiguousSize = contiguousSize;

    p->contiguousRequested = contiguousRequested; /* not a module param. */

    p->devCount = 0;

    for (i = 0; i < gcdLOCAL_MEMORY_COUNT; i++) {
        p->externalBase[i]  = externalBase[i];
        p->externalSize[i]  = externalSize[i];
        p->exclusiveBase[i] = exclusiveBase[i];
        p->exclusiveSize[i] = exclusiveSize[i];
    }

    for (i = 0; i < gcdDEVICE_COUNT; i++) {
        p->platformIDs[i] = platformIDs[i];
        p->hwDevCounts[i] = hwDevCounts[i];
        p->devMemIDs[i] = devMemIDs[i];
        p->devSysMemIDs[i] = devSysMemIDs[i];
        p->devSRAMIDs[i] = devSRAMIDs[i];
        p->devCoreCounts[i] = devCoreCounts[i];
        p->dev2DCoreCounts[i] = dev2DCoreCounts[i];
        if (p->devCoreCounts[i])
            p->devCount++;
    }

    for (i = 0; i < gcdDEVICE_COUNT; i++) {
        for (j = 0; j < SRAM_NUM_PER_DEVICE; j++) {
            p->extSRAMBases[i][j] = extSRAMBases[ i % SRAM_NUM_PER_DEVICE];
            p->extSRAMSizes[i][j] = extSRAMSizes[i % SRAM_NUM_PER_DEVICE];
#if USE_LINUX_PCIE
            p->sRAMBars[i][j] = sRAMBars[j % SRAM_NUM_PER_DEVICE];
            p->sRAMOffsets[i][j] = sRAMOffsets[j % SRAM_NUM_PER_DEVICE];
#endif
        }
    }

    for (i = 0; i < gcdMAX_MAJOR_CORE_COUNT; i++)
        p->userClusterMasks[i] = userClusterMasks[i];

    p->sRAMRequested = sRAMRequested;
    p->sRAMLoopMode  = sRAMLoopMode;

    p->baseAddress = baseAddress;
    p->physSize    = physSize;
    p->bankSize    = bankSize; /* not a module param. */

    p->recovery = recovery;
    p->powerManagement = powerManagement;

    p->softReset = softReset;

    p->enableMmu = mmu;
    p->fastClear = fastClear;

    p->compression = (compression == -1) ?
                     gcvCOMPRESSION_OPTION_DEFAULT : (gceCOMPRESSION_OPTION)compression;
    p->gpu3DMinClock = gpu3DMinClock; /* not a module param. */
    p->enableNN      = enableNN;
    p->registerAPB   = registerAPB;
    p->smallBatch    = smallBatch;

    p->stuckDump  = stuckDump;

    p->deviceType = type;
    p->showArgs   = showArgs;

    p->mmuPageTablePool = mmuPageTablePool;
    p->mmuCmdPool       = mmuCmdPool;
    p->sharedPageTable  = sharedPageTable;
    p->processPageTable = processPageTable;
    p->flatMapping      = flatMapping;

    p->mmuDynamicMap = mmuDynamicMap;
    p->allMapInOne   = allMapInOne;

    p->isrPoll = isrPoll;
#if !gcdENABLE_3D
    p->irqs[0] = -1;
    irqLine = -1;
    p->registerBases[0] = 0;
    registerMemBase     = 0;
    p->registerSizes[0] = 0;
    registerMemSize     = 0;
#endif

#if !gcdENABLE_2D
    p->irq2Ds[0] = -1;
    irqLine2D = -1;
    p->register2DBases[0] = 0;
    registerMemBase2D     = 0;
    p->register2DSizes[0] = 0;
    registerMemSize2D     = 0;
#endif

    p->irqVG = -1;
    irqLineVG = -1;
    p->registerVGBase = 0;
    registerMemBaseVG = 0;
    p->registerVGSize = 0;
    registerMemSizeVG = 0;
}

void
gckOS_DumpParam(void)
{
    gctINT i;

    pr_warn("Galcore options:\n");

#if gcdDEC_ENABLE_AHB
    pr_warn("  registerMemBaseDEC300 = 0x%08lX\n", registerMemBaseDEC300);
    pr_warn("  registerMemSizeDEC300 = 0x%08lX\n", registerMemSizeDEC300);
#endif

    pr_warn("  bankSize          = 0x%08lX\n", bankSize);
    pr_warn("  fastClear         = %d\n",      fastClear);
    pr_warn("  compression       = %d\n",      compression);
    pr_warn("  powerManagement   = %d\n",      powerManagement);
    pr_warn("  baseAddress       = 0x%08lX\n", baseAddress);
    pr_warn("  physSize          = 0x%08lX\n", physSize);
    pr_warn("  recovery          = %d\n",      recovery);
    pr_warn("  stuckDump         = %d\n",      stuckDump);
    pr_warn("  GPU smallBatch    = %d\n",      smallBatch);
    pr_warn("  allMapInOne       = %d\n",      allMapInOne);
    pr_warn("  enableNN          = 0x%x\n",    enableNN);

    pr_warn("  userClusterMasks  = ");
    for (i = 0; i < gcdMAX_MAJOR_CORE_COUNT; i++)
        pr_warn("%x, ", userClusterMasks[i]);
    pr_warn("\n");

    pr_warn("  irqs              = ");
    for (i = 0; i < gcdGLOBAL_CORE_COUNT; i++) {
        if (irqs[i] != -1)
            pr_warn("%d, ", irqs[i]);
    }

    if (irqLine2D != -1)
        pr_warn("  irq2Ds            = ");

    for (i = 0; i < gcdGLOBAL_2D_COUNT; i++) {
        if (irq2Ds[i] != -1)
            pr_warn("%d, ", irq2Ds[i]);
    }
    pr_warn("\n");

    if (irqLineVG != -1) {
        pr_warn("  irqVG: ");
        pr_warn("%d, ", irqLineVG);
        pr_warn("\n");
    }

#if USE_LINUX_PCIE
    pr_warn("Bars configuration: \n");

    for (i = 0; i < gcdGLOBAL_CORE_COUNT; i++) {
        if (bars[i] != -1) {
            pr_warn("  bars[%d] = %d, regOffsets[%d] = %x, ",
                    i, bars[i], i, regOffsets[i]);
            pr_warn("\n");
        }
    }
#endif

    pr_warn("System reserve memory configuration: \n");

    for (i = 0; i < gcdSYSTEM_RESERVE_COUNT; i++) {
        if (contiguousSizes[i]) {
            pr_warn("  contiguousSizes[%d] = 0x%lx\n", i, contiguousSizes[i]);
            pr_warn("  contiguousBases[%d] = 0x%llX\n", i, contiguousBases[i]);
        }
    }

    pr_warn("Registers configuration: \n");

    for (i = 0; i < gcdGLOBAL_CORE_COUNT; i++) {
        if (registerSizes[i]) {
            pr_warn("  bases[%d] = 0x%llx, sizes[%d] = 0x%lx, ",
                    i, registerBases[i], i, registerSizes[i]);
            pr_warn("\n");
        }
    }

    for (i = 0; i < gcdCORE_2D_COUNT; i++) {
        if (register2DSizes[i]) {
            pr_warn("  2D bases[%d] = 0x%llx, sizes[%d] = 0x%lx, ",
                    i, register2DBases[i], i, register2DSizes[i]);
            pr_warn("\n");
        }
    }

    pr_warn("  External sRAMBases = ");
    for (i = 0; i < gcvSRAM_EXT_COUNT; i++)
        pr_warn("0x%llx, ", extSRAMBases[i]);
    pr_warn("\n");

    pr_warn("  mmu               = %d\n", mmu);
    pr_warn("  mmuPageTablePool  = %d\n", mmuPageTablePool);
    pr_warn("  mmuCmdPool        = %d\n", mmuCmdPool);
    pr_warn("  mmuDynamicMap     = %d\n", mmuDynamicMap);
    pr_warn("  isrPoll           = 0x%08X\n", isrPoll);

    pr_warn("Build options:\n");
    pr_warn("  gcdGPU_TIMEOUT    = %d\n", gcdGPU_TIMEOUT);
    pr_warn("  gcdGPU_2D_TIMEOUT = %d\n", gcdGPU_2D_TIMEOUT);
    pr_warn("  gcdINTERRUPT_STATISTIC = %d\n", gcdINTERRUPT_STATISTIC);
}

static int drv_open(struct sn_tranx_t *tdev, struct file *filp)
{
    gceSTATUS status = gcvSTATUS_OK;
    gcsHAL_PRIVATE_DATA_PTR data = gcvNULL;
    gctUINT i, devIndex;
    gctUINT attached = 0;
    gctUINT index    = 0;
    gckDEVICE device;
    struct gpu_t *tgpu = tdev->modules[SN_MODULE_GPU];
    gckGALDEVICE galDevice = tgpu->galDevice;

    gcmkHEADER_ARG("tdev=%p filp=%p", tdev, filp);

    data = kmalloc(sizeof(gcsHAL_PRIVATE_DATA), GFP_KERNEL | __GFP_NOWARN);
    if (data == gcvNULL) {
        gcmkFOOTER_ARG("status=%d", gcvSTATUS_OUT_OF_MEMORY);
        return -ENOMEM;
    }

    data->isLocked = gcvFALSE;
    data->device = galDevice;
    data->pidOpen = task_tgid_vnr(current);

    filp->private_data = data;
    pr_debug("data->pidOpen = %d.\n", data->pidOpen);

    for (devIndex = 0; devIndex < galDevice->args.devCount; devIndex++) {
        device = galDevice->devices[devIndex];

        /* Attached the process. */
        for (i = 0; i < gcvCORE_COUNT; i++) {
            if (device->kernels[i]) {
                status = gckKERNEL_AttachProcess(device->kernels[i], gcvTRUE);

                if (gcmIS_ERROR(status))
                    goto OnError;

                attached = i;
            }
        }
        index = devIndex;
    }

    /* Success. */
    gcmkFOOTER_NO();
    return 0;

OnError:
    for (i = 0; i < attached; i++) {
        if (device->kernels[i])
            gcmkVERIFY_OK(gckKERNEL_AttachProcess(device->kernels[i], gcvFALSE));
    }

    for (devIndex = 0; devIndex < index; devIndex++) {
        device = galDevice->devices[devIndex];

        for (i = 0; i < gcvCORE_COUNT; i++) {
            if (device->kernels[i])
                gcmkVERIFY_OK(gckKERNEL_AttachProcess(device->kernels[i], gcvFALSE));
        }
    }

    kfree(data);

    gcmkFOOTER_ARG("status=%d", status);
    return -ENOTTY;
}

int sn_gpu_open(
    struct sn_tranx_t *tdev,
    struct file* filp
    )
{
    return drv_open(tdev, filp);
}

static int drv_release(struct sn_tranx_t *tdev, struct file *filp)
{
    int ret = -ENOTTY;
    gceSTATUS status = gcvSTATUS_OK;
    gcsHAL_PRIVATE_DATA_PTR data;
    gckGALDEVICE gal_device = gcvNULL;
    gckDEVICE device = gcvNULL;
    gctUINT i, devIndex;

    gcmkHEADER_ARG(" filp=%p", filp);

    data = filp->private_data;
    if (!data) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
		       "%s(%d): private_data is NULL\n",
                       __func__, __LINE__);

        gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
    }

    gal_device = data->device;
    if (!gal_device) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                       "%s(%d): device is NULL\n",
                       __func__, __LINE__);

        gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
    }

    if (data->devIndex >= gcdDEVICE_COUNT) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                       "The galdevice index %d exceed devicecount %d.\n",
                       data->devIndex, gcdDEVICE_COUNT);

        gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
    }

    if (data->isLocked) {
        device = gal_device->devices[data->devIndex];

        /* Release the mutex. */
        gcmkONERROR(gckOS_ReleaseMutex(device->os, device->commitMutex));
        data->isLocked = gcvFALSE;
    }

    /* A process gets detached. */
    for (devIndex = 0; devIndex < gal_device->args.devCount; devIndex++) {
        device = gal_device->devices[devIndex];

        /* A process gets detached. */
        for (i = 0; i < gcvCORE_COUNT; i++) {
            if (device->kernels[i]) {
                gcmkVERIFY_OK(gckKERNEL_AttachProcessEx(device->kernels[i],
                                                        gcvFALSE, data->pidOpen));
            }
        }
    }
    gc_hal_osal_remove_mapping(data->pidOpen);

    kfree(data);
    filp->private_data = NULL;

    /* Success. */
    ret = 0;

OnError:
    gcmkFOOTER();
    return ret;
}

int sn_gpu_release(
    struct sn_tranx_t *tdev,
    struct file* filp
    )
{
    return drv_release(tdev, filp);
}

#if (SUB_SYS_VIP || SUB_SYS_GPU)
static int gpu_keep_soft_reset(struct sn_tranx_t *tdev, enum SYS_CON_SUBSYS subsys)
{
    switch(subsys){
        case SYS_CTL_VIP:
#if SUB_SYS_VIP
            sys_config_reset_keep(tdev, 0, subsys);
            sys_config_reset_keep(tdev, 1, subsys);
#endif
            break;
        case SYS_CTL_GPU2D:
#if SUB_SYS_GPU
#if S1_GPU
            sys_config_reset_keep(tdev, 0, subsys);
#endif

#if S2_GPU
            sys_config_reset_keep(tdev, 1, subsys);
#endif
#endif
            break;
        default:
            sn_pri(tdev, SN_INF, "%s, subsys incorrect\n", __func__);
            break;
    }
    return 0;
}


static int gpu_release_soft_reset(struct sn_tranx_t *tdev, enum SYS_CON_SUBSYS subsys)
{
    switch(subsys){
        case SYS_CTL_VIP:
#if SUB_SYS_VIP
            sys_config_reset_release(tdev, 0, subsys);
            sys_config_reset_release(tdev, 1, subsys);
#endif
            break;
        case SYS_CTL_GPU2D:
#if SUB_SYS_GPU
#if S1_GPU
            sys_config_reset_release(tdev, 0, subsys);
#endif

#if S2_GPU
            sys_config_reset_release(tdev, 1, subsys);
#endif
#endif
            break;
        default:
            sn_pri(tdev, SN_INF, "%s, subsys incorrect\n", __func__);
            break;
    }
    return 0;
}
#endif

static int gpu_soft_reset(struct sn_tranx_t *tdev, enum SYS_CON_SUBSYS subsys)
{
    struct gpu_t *tgpu = tdev->modules[SN_MODULE_GPU];
    gckDEVICE device = gcvNULL;
    gctUINT i, devIndex;

    switch(subsys){
        case SYS_CTL_VIP:
#if SUB_SYS_VIP
            gpu_keep_soft_reset(tdev, subsys);
            gpu_release_soft_reset(tdev, subsys);
#endif
            break;
        case SYS_CTL_GPU2D:
#if SUB_SYS_GPU
            gpu_keep_soft_reset(tdev, subsys);
            gpu_release_soft_reset(tdev, subsys);
#endif
            break;
        default:
            sn_pri(tdev, SN_INF, "%s, subsys incorrect\n", __func__);
            break;
    }

    //Soft reset for the GPU HW
    for (devIndex = 0; devIndex < tgpu->galDevice->args.devCount; devIndex++) {
        device = tgpu->galDevice->devices[devIndex];

        for (i = 0; i < gcvCORE_COUNT; i++) {
            switch(subsys) {
                case SYS_CTL_VIP:
#if SUB_SYS_VIP
                    if (device->kernels[i] && i >= gcvCORE_MAJOR && i <= gcvCORE_3D_MAX)
                        gckHARDWARE_Reset(device->kernels[i]->hardware);
#endif
                    break;
                case SYS_CTL_GPU2D:
#if SUB_SYS_GPU
                    if (device->kernels[i] && i >= gcvCORE_2D && i <= gcvCORE_2D_MAX)
                        gckHARDWARE_Reset(device->kernels[i]->hardware);
#endif
                    break;
                default:
                    sn_pri(tdev, SN_INF, "%s, subsys incorrect\n", __func__);
                    break;
            }
        }
    }

    return 0;
}

static long drv_osalctl(struct file *filp, unsigned int ioctlCode, unsigned long arg, struct sn_tranx_t *tdev)
{
    long ret = -ENOTTY;
    gceSTATUS status = gcvSTATUS_OK;
    gcsHAL_INTERFACE *iface;
    gckGALDEVICE gal_device;
    gckDEVICE device;
    gcsHAL_PRIVATE_DATA_PTR data;
    gctUINT64               ep_addr;
    gctUINT32               size;

#if VIVANTE_PROFILER
    static gcsHAL_PROFILER_INTERFACE *iface_profiler;
#endif

    gcmkHEADER_ARG("filp=%p ioctlCode=%u arg=%lu", filp, ioctlCode, arg);

    data = filp->private_data;

    if (data == gcvNULL) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                       "%s(%d): private_data is NULL\n",
                       __func__, __LINE__);

        gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
    }

    gal_device = data->device;

    if (gal_device == gcvNULL) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                       "%s(%d): device is NULL\n",
                       __func__, __LINE__);

        gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
    }

    switch (ioctlCode) {
    case IOCTL_GCHAL_INTERFACE:
        iface = (gcsHAL_INTERFACE *)(void *)arg;
        gcmkTRACE_ZONE(gcvLEVEL_INFO, gcvZONE_DRIVER, "iface->command = %d.\n", iface->command);

        if (iface->command == gcvHAL_DEVICE_MUTEX) {
            if (iface->u.DeviceMutex.isMutexLocked == gcvTRUE)
                data->isLocked = gcvTRUE;
            else
                data->isLocked = gcvFALSE;
        }

        iface->osal_filp = filp;
        if (iface->command == gcvHAL_WRAP_USER_MEMORY) {
            ep_addr = iface->u.WrapUserMemory.desc.physical;
            gcmkTRACE_ZONE(gcvLEVEL_INFO, gcvZONE_DRIVER,"ep_addr = 0x%llx.\n", ep_addr);
            size = iface->u.WrapUserMemory.desc.size;
            iface->u.WrapUserMemory.desc.physical =
                sn_mem_osal_translate_handle(tdev, filp, ep_addr, size);
            if (ep_addr == 1) // invalid
                gcmONERROR(gcvSTATUS_INVALID_ARGUMENT);

        }

        /* Record the last device id. */
        data->devIndex = iface->devIndex;

        if (iface->command == gcvHAL_CHIP_INFO) {
            gctUINT i, count = 0;

            for (i = 0; i < gal_device->args.devCount; i++) {
                device = gal_device->devices[i];

                gcmkONERROR(gckDEVICE_ChipInfo(device, iface, &count));

                /* For same open process, keep same task-id and file hanlder. */
                if (device->kernels[i]) {
                    gckKERNEL_SetProcess_TaskId(device->kernels[i], data->pidOpen,
                                                        iface->osal_task_id, iface->osal_filp);
                }
            }
        } else {
            if (iface->devIndex >= gcdDEVICE_COUNT)
                gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);

            device = gal_device->devices[iface->devIndex];

            status = gckDEVICE_Dispatch(device, iface);

            /* Redo system call after pending signal is handled. */
            if (status == gcvSTATUS_INTERRUPTED) {
                ret = -ERESTARTSYS;
                gcmkONERROR(status);
            }
        }

        break;

    case IOCTL_GCHAL_PROFILER_INTERFACE:
#if VIVANTE_PROFILER
        iface_profiler = (gcsHAL_PROFILER_INTERFACE *)(void *)arg;

        if (iface_profiler->devIndex >= gcdDEVICE_COUNT)
            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);

        device = gal_device->devices[iface_profiler->devIndex];

        status = gckDEVICE_Profiler_Dispatch(device, iface_profiler);

        /* Redo system call after pending signal is handled. */
        if (status == gcvSTATUS_INTERRUPTED) {
            ret = -ERESTARTSYS;
            gcmkONERROR(status);
        }

#endif
        break;

    default:
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                       "%s(%d): unknown command %d\n",
                       __func__, __LINE__, ioctlCode);

        gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
    }

    /* Success. */
    ret = 0;

OnError:
    gcmkFOOTER();
    return ret;
}

static long drv_ioctl(struct file *filp, unsigned int ioctlCode, unsigned long arg, struct sn_tranx_t *tdev)
{
    long ret = -ENOTTY;
    gceSTATUS status = gcvSTATUS_OK;
    gcsHAL_INTERFACE iface;
    gctUINT32 copyLen;
    DRIVER_ARGS drvArgs;
    gckGALDEVICE gal_device;
    gckDEVICE device;
    gcsHAL_PRIVATE_DATA_PTR data;
#if VIVANTE_PROFILER
    static gcsHAL_PROFILER_INTERFACE iface_profiler;
#endif

    gcmkHEADER_ARG("filp=%p ioctlCode=%u arg=%lu", filp, ioctlCode, arg);

    data = filp->private_data;

    if (data == gcvNULL) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                       "%s(%d): private_data is NULL\n",
                       __func__, __LINE__);

        gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
    }

    gal_device = data->device;

    if (gal_device == gcvNULL) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                       "%s(%d): device is NULL\n",
                       __func__, __LINE__);

        gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
    }

    switch (ioctlCode) {
    case IOCTL_GCHAL_INTERFACE:
        /* Get the drvArgs. */
        copyLen = copy_from_user(&drvArgs, (void *)arg, sizeof(DRIVER_ARGS));

        if (copyLen != 0) {
            gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                           "%s(%d): error copying of the input arguments.\n",
                           __func__, __LINE__);

            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
        }

        /* Now bring in the gcsHAL_INTERFACE structure. */
        if (drvArgs.InputBufferSize != sizeof(gcsHAL_INTERFACE) ||
            drvArgs.OutputBufferSize != sizeof(gcsHAL_INTERFACE)) {
            gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                           "%s(%d): input or/and output structures are invalid.\n",
                           __func__, __LINE__);

            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
        }

        copyLen = copy_from_user(&iface,
                                 gcmUINT64_TO_PTR(drvArgs.InputBuffer),
                                 sizeof(gcsHAL_INTERFACE));

        if (copyLen != 0) {
            gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                           "%s(%d): error copying of input HAL interface.\n",
                           __func__, __LINE__);

            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
        }

        /* assign osal file handler */
        iface.osal_filp = filp;

        if (iface.command == gcvHAL_DEVICE_MUTEX) {
            if (iface.u.DeviceMutex.isMutexLocked == gcvTRUE)
                data->isLocked = gcvTRUE;
            else
                data->isLocked = gcvFALSE;
        }

        /* Record the last device id. */
        data->devIndex = iface.devIndex;

        if (iface.command == gcvHAL_CHIP_INFO) {
            gctUINT i, count = 0;

            for (i = 0; i < gal_device->args.devCount; i++) {
                device = gal_device->devices[i];

                gcmkONERROR(gckDEVICE_ChipInfo(device, &iface, &count));
            }
        } else {
            if (iface.devIndex >= gcdDEVICE_COUNT)
                gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);

            device = gal_device->devices[iface.devIndex];

            status = gckDEVICE_Dispatch(device, &iface);

            /* Redo system call after pending signal is handled. */
            if (status == gcvSTATUS_INTERRUPTED) {
                ret = -ERESTARTSYS;
                gcmkONERROR(status);
            }
        }

        /* Copy data back to the user. */
        copyLen = copy_to_user(gcmUINT64_TO_PTR(drvArgs.OutputBuffer),
                               &iface,
                               sizeof(gcsHAL_INTERFACE));

        if (copyLen != 0) {
            gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                           "%s(%d): error copying of output HAL interface.\n",
                           __func__, __LINE__);

            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
        }
        break;

    case IOCTL_GCHAL_PROFILER_INTERFACE:
#if VIVANTE_PROFILER
        /* Get the drvArgs. */
        copyLen = copy_from_user(&drvArgs, (void *)arg, sizeof(DRIVER_ARGS));

        if (copyLen != 0) {
            gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                           "%s(%d): error copying of the input arguments.\n",
                           __func__, __LINE__);

            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
        }

        /* Now bring in the gcsHAL_INTERFACE structure. */
        if (drvArgs.InputBufferSize != sizeof(gcsHAL_PROFILER_INTERFACE) ||
            drvArgs.OutputBufferSize != sizeof(gcsHAL_PROFILER_INTERFACE)) {
            gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                           "%s(%d): input or/and output structures are invalid.\n",
                           __func__, __LINE__);

            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
        }

        copyLen = copy_from_user(&iface_profiler,
                                 gcmUINT64_TO_PTR(drvArgs.InputBuffer),
                                 sizeof(gcsHAL_PROFILER_INTERFACE));

        if (copyLen != 0) {
            gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                           "%s(%d): error copying of input HAL interface.\n",
                           __func__, __LINE__);

            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
        }

        if (iface_profiler.devIndex >= gcdDEVICE_COUNT)
            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);

        device = gal_device->devices[iface_profiler.devIndex];

        status = gckDEVICE_Profiler_Dispatch(device, &iface_profiler);

        /* Redo system call after pending signal is handled. */
        if (status == gcvSTATUS_INTERRUPTED) {
            ret = -ERESTARTSYS;
            gcmkONERROR(status);
        }

        /* Copy data back to the user. */
        copyLen = copy_to_user(gcmUINT64_TO_PTR(drvArgs.OutputBuffer),
                               &iface_profiler,
                               sizeof(gcsHAL_PROFILER_INTERFACE));

        if (copyLen != 0) {
            gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                           "%s(%d): error copying of output HAL interface.\n",
                           __func__, __LINE__);

            gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
        }
#endif
        break;

    case IOCTL_GCHAL_VIP_SOFT_RESET:
        gpu_soft_reset(tdev, SYS_CTL_VIP);
        break;

    case IOCTL_GCHAL_GPU2D_SOFT_RESET:
        gpu_soft_reset(tdev, SYS_CTL_GPU2D);
        break;

    default:
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                       "%s(%d): unknown command %d\n",
                       __func__, __LINE__, ioctlCode);

        gcmkONERROR(gcvSTATUS_INVALID_ARGUMENT);
    }

    /* Success. */
    ret = 0;

OnError:
    gcmkFOOTER();
    return ret;
}

static void osal_gpu_handler(struct work_struct *workPtr) {
    sn_osal_work* work = (sn_osal_work*) workPtr;
    osal_command* command    = work->cmdList.cmds;
    osal_response* response  = work->rspList.rsps;
    struct sn_tranx_t *tdev  = work->tdev;
    struct file *filp = NULL;
    gcsHAL_PRIVATE_DATA_PTR data = gcvNULL;
    gctUINT i;
    unsigned int cmdcode = 0;
    gcsHAL_INTERFACE iface;
    int InputSize = 0;
    long ret = 0;
    gctUINT32 kpid;

    filp = work->filp;
    if (filp == NULL) {
        sn_pri(tdev, SN_ERR, "%s: filp is NULL.\n", __func__);
        return;
    }

    data = filp->private_data;
    gckOS_GetThreadID(&kpid);
    if (data->pidOpen != kpid)
        gc_hal_osal_add_mapping(kpid, data->pidOpen);

    gcmkTRACE_ZONE(gcvLEVEL_INFO, gcvZONE_DRIVER,"work->cmdList.num_cmds = %d.\n", work->cmdList.num_cmds);
    for (i = 0; i < work->cmdList.num_cmds; ++i, ++command, ++response) {
        gcmkTRACE_ZONE(gcvLEVEL_INFO, gcvZONE_DRIVER, "%s: work_queue recvd cmd %d\n", __func__, command->header.op_code);
        gcmkTRACE_ZONE(gcvLEVEL_INFO, gcvZONE_DRIVER, "%s: work_queue cmd data words %d\n", __func__, command->header.num_data_words);
        InputSize = command->header.num_data_words;
        switch (command->header.op_code) {
            case OSAL_ACCL_CMD_GPU_RUN:
                cmdcode = IOCTL_GCHAL_INTERFACE;
                memcpy(&iface, command->data, InputSize);
                ret = drv_osalctl(filp, cmdcode, (uintptr_t)&iface, tdev);
                memcpy(response->data, &iface, InputSize);
               break;
            case OSAL_ACCL_CMD_GPU_PROFILER_RUN:
                cmdcode = IOCTL_GCHAL_PROFILER_INTERFACE;
                memcpy(&iface, command->data, InputSize);
                ret = drv_osalctl(filp, cmdcode, (uintptr_t)&iface, tdev);
                memcpy(response->data, &iface, InputSize);
                break;
            default:
                sn_pri(tdev, SN_ERR, "%s: invalid cmd %d.\n", __func__, command->header.op_code);
                work->rspList.num_rsps = 1;
                response->data[0] = 0xdeadbeaf;
                break;
        }
    }

    if (ret < 0)
        sn_pri(tdev, SN_ERR, "%s: failed.\n");

    sn_osal_finish_work(work);

    gc_hal_osal_remove_single_mapping(kpid);
}

static void osal_gpu_close(struct sn_tranx_t* tdev, struct file* filp)
{
    struct gpu_t *tgpu = tdev->modules[SN_MODULE_GPU];
    gckDEVICE device = gcvNULL;
    gctUINT i, devIndex;

    for (devIndex = 0; devIndex < tgpu->galDevice->args.devCount; devIndex++) {
        device = tgpu->galDevice->devices[devIndex];
        for (i = 0; i < gcvCORE_COUNT; i++) {
            if (device->kernels[i])
                gckOS_SignalAll(device->kernels[i]->os);
        }
    }
}

long sn_gpu_ioctl(
    struct file* filp,
    unsigned int ioctlCode,
    unsigned long arg,
    struct sn_tranx_t *tdev
    )
{
    return drv_ioctl(filp, ioctlCode, arg, tdev);
}

static uint32_t getLoad(uint32_t total, uint32_t idle, uint32_t* prev_total, uint32_t* prev_idle)
{
    uint32_t total_cycle = total - *prev_total;
    uint32_t idle_cycle = idle - *prev_idle;
    uint32_t load;
    if (idle_cycle > total_cycle) { // The values are not read together atomically.  When idle, this happens.
        total_cycle = idle_cycle;
    }
    load = (total_cycle - idle_cycle) * 100ull / total_cycle;
    *prev_total = total;
    *prev_idle = idle;
    return load;
}

static void tgpu_loading_timer_isr(struct timer_list *t)
{
    struct gpu_t *tgpu = from_timer(tgpu, t, loading_timer);
    uint32_t load;
    int i;
    gckDEVICE gckdevice = gcvNULL;

    if (!tgpu || !tgpu->galDevice) {
        pr_warn("the gpu device is not ready.");
        mod_timer(&tgpu->loading_timer, jiffies + LOADING_TIME * HZ);
        return;
    }
    gckdevice = tgpu->galDevice->devices[0];

    for (i = 0; gckdevice->kernels[i] && gckdevice->kernels[i]->hardware; ++i) {
        uint32_t total_cycle, idle_cycle;
        gckHARDWARE_QueryCycleCount(gckdevice->kernels[i]->hardware, &total_cycle, &idle_cycle);
        load = getLoad(total_cycle, idle_cycle, tgpu->last_total_cycle + i, tgpu->last_idle_cycle + i);
        sn_perf_load(tgpu->vip_perf_handle, i, load);
    }

    for (i = 0; gckdevice->map[gcvHARDWARE_2D].kernels[i] && gckdevice->map[gcvHARDWARE_2D].kernels[i]->hardware; ++i) {
        uint32_t total_cycle, idle_cycle;
        gckHARDWARE_QueryCycleCount(gckdevice->map[gcvHARDWARE_2D].kernels[i]->hardware, &total_cycle, &idle_cycle);
        load = getLoad(total_cycle, idle_cycle, tgpu->last_total_cycle + 4 + i, tgpu->last_idle_cycle + 4 + i);
        sn_perf_load(tgpu->gc_perf_handle, i, load);
    }
    //restart timer
    mod_timer(&tgpu->loading_timer, t->expires + LOADING_TIME * HZ);
}

static int drv_init(struct gpu_t *tgpu)
{
    int result = -EINVAL;
    gceSTATUS status;
    gckGALDEVICE device = gcvNULL;
    gcsPLATFORM *platform = tgpu->platform;
    //struct class *device_class = gcvNULL;

    gcmkHEADER();

    pr_info("Galcore version %s\n", gcvVERSION_STRING);

    if (showArgs)
        gckOS_DumpParam();

    if (!platform) {
	    pr_info("failed to get params.\n");
	    return -1;
    }

    /* Create the GAL device. */
    status = gckGALDEVICE_Construct(platform, &platform->params, &device);

    if (gcmIS_ERROR(status)) {
        gcmkTRACE_ZONE(gcvLEVEL_ERROR, gcvZONE_DRIVER,
                       "%s(%d): Failed to create the GAL device: status=%d\n",
                       __func__, __LINE__, status);

        goto OnError;
    }

    /* Start the GAL device. */
    gcmkONERROR(gckGALDEVICE_Start(device));

    /* Set global galDevice pointer. */
    tgpu->galDevice = device;

    /* Success. */
    gcmkFOOTER();
    return 0;

OnError:

    if (device)
    {
        gcmkVERIFY_OK(gckGALDEVICE_Stop(device));
        gcmkVERIFY_OK(gckGALDEVICE_Destroy(device));
    }

    gcmkFOOTER();
    return result;
}

static void drv_exit(struct gpu_t *tgpu)
{
    gcmkHEADER();

    del_timer_sync(&tgpu->loading_timer);

    gcmkVERIFY_OK(gckGALDEVICE_Stop(tgpu->galDevice));
    gcmkVERIFY_OK(gckGALDEVICE_Destroy(tgpu->galDevice));

    gcmkFOOTER_NO();
}

int sys_gpu_init(struct sn_tranx_t *tdev) {
    int ret = -ENODEV;
    struct gpu_t *tgpu;
    gcsPLATFORM *platform = NULL;

    tgpu = kzalloc(sizeof(struct gpu_t), GFP_KERNEL);
    if (!tgpu) {
        sn_pri(tdev, SN_ERR, "gpu: kmalloc gpu_t failed\n");
        return -ENOMEM;
    }

    tdev->modules[SN_MODULE_GPU] = tgpu;

    gckPLATFORM_Init(tdev->pdev, &tgpu->platform);
    platform = tgpu->platform;
    platform->dev = &tdev->pdev->dev;
    platform->name = tdev->dev_name;

    switch(tdev->pf_vf_mode)
    {
        case PF_MODE:
            memcpy(regOffsets, PF_regOffsets, sizeof(regOffsets));
            break;
        case TWO_VF_MODE:
            if(tdev->vf_index == VF2_INDEX) {
                memcpy(regOffsets, VF2_regOffsets, sizeof(regOffsets));
            }
            else if(tdev->vf_index == VF1_INDEX) {
                memcpy(regOffsets, VF1_regOffsets, sizeof(regOffsets));
            }
            break;
        case ONE_VF_MODE:
            memcpy(regOffsets, VF_regOffsets, sizeof(regOffsets));
            break;
        default:
            sn_pri(tdev, SN_INF, "%s,vf_max_count incorrect\n", __func__);
            break;
    }

    /* Gather module parameters. */
    _InitModuleParam(&platform->params);
    platform->params.devices[0] = &tdev->pdev->dev;

    if (platform->ops->adjustParam) {
        /* Override default module param. */
        sn_pri(tdev, SN_DBG, "gpu: adjust params.\n");
        platform->ops->adjustParam(platform, &platform->params);
    }

    /* Update module param because drv_init() uses them directly. */
    /* _SyncModuleParam(&platform->params); */

    if (powerManagement == 0)
        gcmkPRINT("[galcore warning]: power saving is disabled.");

    if (debugLevel != 0) {
        gckOS_SetDebugLevelZone(gcvLEVEL_VERBOSE, gcdZONE_ALL);
    }

    ret = drv_init(tgpu);

    if (ret < 0) {
        sn_pri(tdev, SN_ERR, "gpu: drv_init failed\n");
        goto out_free_tgpu;
    }

    sn_osal_register(tdev, "gpu2d", osal_gpu2d, osal_gpu_handler, osal_gpu_close);
    sn_osal_register(tdev, "vip", osal_ml, osal_gpu_handler, osal_gpu_close);
    sn_pri(tdev, SN_INF, "gpu: osal handler registered.\n");

    tgpu->vip_perf_handle = sn_perf_register(tdev, "vip", SN_PERF_ID_VIP, NULL, 4);
    tgpu->gc_perf_handle = sn_perf_register(tdev, "gc620", SN_PERF_ID_GC620, NULL, 2);
    timer_setup(&tgpu->loading_timer, tgpu_loading_timer_isr, 0);
    mod_timer(&tgpu->loading_timer, jiffies);

    sn_pri(tdev, SN_ERR, "gpu: module inserted\n");
    return ret;

out_free_tgpu:
    gckPLATFORM_Terminate(platform);
    kfree(tgpu);
    sn_pri(tdev, SN_ERR, "gpu: module not inserted\n");
    return ret;
}

void sys_gpu_exit(struct sn_tranx_t *tdev) {
    struct gpu_t *tgpu = tdev->modules[SN_MODULE_GPU];
    gcsPLATFORM *platform = tgpu->platform;

    drv_exit(tgpu);
    gckPLATFORM_Terminate(platform);

    kfree(tgpu);
    sn_pri(tdev, SN_DBG, "gpu: remove module done\n");
}
