
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <hugetlbfs.h>

#include "transcoder.h"

#define DIRECTION_RC2EP 0
#define DIRECTION_EP2RC 1

static int usage(char **argv)
{
	printf("usage:\n");
	printf("\t%s device_node transfer_direction test_count size_MB\n",
	       argv[0]);
	printf("\tdevice_node       : /dev/ama_transcoderN\n");
	printf("\ttransfer_direction: rc2ep | ep2rc\n");
	printf("\ttest_count        : transfer times\n");
	printf("\tsize              : transfer data size in MB(1~64)\n");
	printf("example:\n");
	printf("\t%s /dev/ama_transcoder0 rc2ep ep_addr 1 32 ---> rc2ep transfer 1 times data, each data size is 32MB with /dev/ama_transcoder0\n",
	       argv[0]);
	printf("\t%s /dev/ama_transcoder1 ep2rc ep_addr 2 64 ---> ep2rc transfer 2 times data, each data size is 64MB with /dev/ama_transcoder1\n",
	       argv[0]);
}

int main(int argc, char **argv)
{
	int ret = -1;
	int fd;
	int direction = -1; //0: rc2ep 1:ep2rc
	unsigned int poll_intr = 1; // interrupt mode
	unsigned long rc_addr, ep_addr;
	unsigned long time_cnt = 0, diff = 0;
	unsigned long src, src_h, src_l, dst, dst_h, dst_l, size, i;
	struct trans_pcie_hdma hdma_trans;
	struct timeval time, time_1;
	unsigned int count = 0;
	float bw = 0;

	do {
		if (argc != 6) {
			ret = -2;
			break;
		}
		if (strncmp(argv[2], "rc2ep", 5) == 0) {
			direction = DIRECTION_RC2EP;
		} else if (strncmp(argv[2], "ep2rc", 5) == 0) {
			direction = DIRECTION_EP2RC;
		} else {
			printf("[ERROR] please input correct transfer direction param, invalid param: %s\n",
			       argv[2]);
			ret = -3;
			break;
		}
		count = strtoul(argv[4], 0, 0);
		if (count <= 0) {
			printf("[ERROR] please input correct count number param, invalid param: %s\n",
			       argv[4]);
			ret = -4;
			break;
		}
		size = strtoul(argv[5], 0, 0);
		if (size <= 0) {
			printf("[ERROR] please input correct data size param, invalid param: %s\n",
			       argv[5]);
			ret = -5;
			break;
		}
		if (size > 64)
			size = 64;
		if (size <= 0)
			size = 32;
		size = size * 1024 * 1024;
		fd = open(argv[1], O_RDWR, S_IRUSR | S_IWUSR);
		if (fd == -1) {
			printf("[ERROR] failed to open: %s, please check the device node is exist...\n",
			       argv[1]);
			break;
		}
		printf("device node is %s \n", argv[1]);
		ret = 0;
	} while (0);
	if (ret != 0) {
		usage(argv);
		exit(-1);
	}

	do {
		rc_addr = (unsigned long)get_huge_pages(size, GHP_DEFAULT);
		if (rc_addr == 0) {
			printf("get_huge_pages failed.\n");
			ret = -1;
			break;
		}

		ep_addr = strtoul(argv[3], 0, 0);
		printf("EP address=0x%lx \n", ep_addr);

		if (DIRECTION_RC2EP == direction) {
			src = rc_addr;
			dst = ep_addr; // can fix a special addr.
			src_l = src & 0xffffffff;
			src_h = src >> 32;
			dst_l = dst & 0xffffffff;
			dst_h = dst >> 32;

			hdma_trans.direct = RC2EP;
			hdma_trans.interrupt_enable =
				poll_intr; //INTE_EN;  POLL_EN
			hdma_trans.size = size;
			hdma_trans.sar_high = src_h;
			hdma_trans.sar_low = src_l;
			hdma_trans.dar_high = 0x0;
			hdma_trans.dar_low = dst;

		} else {
			dst = rc_addr;
			src = ep_addr; // can fix a special addr.
			src_l = src & 0xffffffff;
			src_h = src >> 32;
			dst_l = dst & 0xffffffff;
			dst_h = dst >> 32;

			hdma_trans.direct = EP2RC;
			hdma_trans.interrupt_enable =
				poll_intr; //INTE_EN;  POLL_EN
			hdma_trans.size = size;
			hdma_trans.sar_high = 0x0;
			hdma_trans.sar_low = src;
			hdma_trans.dar_high = dst_h;
			hdma_trans.dar_low = dst_l;
		}
		printf("normal hdma trans size=0x%x, count=%d..\n", size,
		       count);
		for (i = 0; i < count; i++) {
			gettimeofday(&time, NULL);
			if (ret = ioctl(fd, CB_TRANX_HDMA_TRANX_VIR,
					&hdma_trans)) //enable hdma transmit
			{
				printf("hdma tranx failed,ret=0x%x", ret);
				goto out;
			}
			gettimeofday(&time_1, NULL);
			diff = 1000000 * (time_1.tv_sec - time.tv_sec) +
			       time_1.tv_usec - time.tv_usec;
			time_cnt += diff;
			printf(" round %d/%d time=%ldus.\n", i + 1, count,
			       diff);
		}

		size = size / 1024 / 1024;
		bw = count * size * 1000 * 1000 / time_cnt;
		printf("total_time=%ld us, avg_bw= %f MB/s\n", time_cnt, bw);
	} while (0);

out:
	free_huge_pages((void *)rc_addr);

	close(fd);

	return 0;
}
